/*
  ChaOne ver. 1.1.0
  version as of 17.Aug.2003
  (c) 2001-2003 Atsushi YAMADA (Studio ARC)
  (based on JAXP (SAX, DOM) using Relaxer object)
*/

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Attr;
import org.w3c.dom.Text;
import org.xml.sax.InputSource;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import java.io.*;
import java.util.*;
import java.net.URL;

public class chaone {
    private static String input_file_name = "";
    private static String standalone = "";
    private static String encoding = "";
    private static String version = "1.1.0";
    private static boolean debug = false;

    // Resources
    private static String ea_symbol_file_name = "ea_symbol_table.xml";
    private static String hk_table_file_name = "hk_table.xml";
    private static String chunk_rules_file_name = "chunk_rules.xml";
    private static String ipaform_file_name = "IPAForm.xml";
    private static String fpaform_file_name = "FPAForm.xml";

    static HashMap EASymbolTable = new HashMap();
    static HashMap HKTable = new HashMap();
    static HashMap ChunkRuleIndex = new HashMap();
    static HashMap ChunkRuleW2Index = new HashMap();
    static HashMap ChunkRuleIndexLoc = new HashMap();
    static HashMap ChunkRules = new HashMap();
    static HashMap IPAFormTable = new HashMap();
    static HashMap IPAFormPronTable = new HashMap();
    static HashMap FPAFormTable = new HashMap();
    static HashMap FPAFormPronTable = new HashMap();
    static Document doc;
    
    public static void main(String[] args) throws Exception {

	processArgs(args);
	
	if (encoding.length() == 0) {
	    // set default encoding for OS
	    encoding = System.getProperty("file.encoding");
	    if (encoding.startsWith("EUC-JP")) encoding = "EUC-JP";
	    else if (encoding.equals("MS932")) encoding = "Shift_JIS";
	}

	// read and prepare resources to be used
	prepareEASymbolTable();
	prepareHKTable();
	prepareChunkRules();
	prepareIPAFormTable();
	prepareFPAFormTable();
	
	// get input stream
	InputStream source = (input_file_name.length() > 0 ? new FileInputStream(input_file_name) : System.in);
	StringWriter sWriter = new StringWriter();
	sWriter.write("<?xml version=\"1.0\" encoding=\"" + encoding + "\"?>\n");
	try {
	    BufferedReader bReader = new BufferedReader(new InputStreamReader(source, encoding));
	    String input;
	    while ((input = bReader.readLine()) != null) {
		sWriter.write(input + "\n");
		if (input.equals("</S>")) {
		    DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
		    DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
		    TransformerFactory tFactory = TransformerFactory.newInstance();
		    Transformer xmlOut = tFactory.newTransformer();
		    Properties outputProps = new Properties();
		    outputProps.setProperty("method", "xml");
		    outputProps.setProperty("encoding", encoding);
		    outputProps.setProperty("indent", "yes");
		    outputProps.setProperty("omit-xml-declaration", "yes");
		    xmlOut.setOutputProperties(outputProps);
		    try {
			doc = dBuilder.parse(new InputSource(new StringReader(sWriter.toString())));
			Element root = doc.getDocumentElement();
			if (!(standalone.equals("chaone"))) {
			    // call Chunker routine
			    chaonePreprocess(root);
			    chaoneChunker(root);
			}
			if (!(standalone.equals("chunker"))) {
			    // call ChaOne routine
			    chaonePAlt(root);
			}
			xmlOut.transform(new DOMSource(doc), new StreamResult(System.out));
		    } catch (Exception e) {
			System.out.println("<S></S>");
			if (debug) {
			    System.err.println("Exception: " + e.getMessage());
			    e.printStackTrace();
			}
		    }
		    sWriter.close();
		    // next sentence
		    sWriter = new StringWriter();
		    sWriter.write("<?xml version=\"1.0\" encoding=\"" + encoding + "\"?>\n");
		}
	    }
	    sWriter.close();
	} catch (Exception e) {
	    if (debug) {
		System.err.println("Exception: " + e.getMessage());
		e.printStackTrace();
	    }
	}
    }
    
    public static void processArgs(String[] args) {
	// argment check
	if (args.length > 0) {
	    int i;
	    // for args starting with '-'
	    for (i = 0; i < args.length && args[i].charAt(0) == '-'; i++) {
		if (debug) System.err.println("-arg " + i + ":" + args[i]);
		if (args[i].equals("-h") | args[i].equals("-help"))
		    // help
		    printUsage();
		else if (args[i].equals("-v") | args[i].equals("-version"))
		    // version
		    printVersion();
		else if (args[i].startsWith("-s")) {
		    // standalone switch
		    if (args[i].substring(3).equals("chaone"))
			standalone = "chaone";
		    else if (args[i].substring(3).equals("chunker"))
			standalone = "chunker";
		    else
			printUsage();
		}
		else if (args[i].startsWith("-e")) {
		    // set encoding
		    if (args[i].substring(3).equals("ISO-2022-JP"))
			encoding = "ISO-2022-JP";
		    else if (args[i].substring(3).equals("EUC-JP")
			     | args[i].substring(3).equals("EUC_JP"))
			encoding = "EUC-JP";
		    else if (args[i].substring(3).equals("Shift_JIS")
			     | args[i].substring(3).equals("Shift-JIS")
			     | args[i].substring(3).equals("SJIS"))
			encoding = "Shift_JIS";
		    else
			printUsage();
		}
		else printUsage();
	    }
	    if (args.length - i > 1) printUsage();
	    else if (args.length - i > 0) {
		if (debug) System.err.println("+arg " + i + ":" + args[i]);
		// input from file
		input_file_name = args[i];
		File input_file = new File(input_file_name);
		if (!input_file.exists()) {
		    System.err.println("input file " + input_file_name + " does not exist.");
		    return;
		}
	    }
	}
    }
    
    public static void prepareEASymbolTable() throws Exception {
	// prepare EASymbolTable
	SAXParserFactory spFactory = SAXParserFactory.newInstance();
	SAXParser sParser = spFactory.newSAXParser();
	sParser.parse(chaone.class.getResourceAsStream(ea_symbol_file_name),
		      new EASymbolTableHandler());
    }
    
    public static void prepareHKTable() throws Exception {
	// prepare HKTable
	SAXParserFactory spFactory = SAXParserFactory.newInstance();
	SAXParser sParser = spFactory.newSAXParser();
	sParser.parse(chaone.class.getResourceAsStream(hk_table_file_name),
		      new HKTableHandler());
    }

    public static void prepareChunkRules() throws Exception {
	// prepare ChunkRules using Relaxer object
	ChunkRules chunk_rules = new ChunkRules(chaone.class.getResourceAsStream(chunk_rules_file_name));
	Rule[] rules = chunk_rules.getRule();
	for (int i = 0; i < rules.length; i++) {
	    String id = rules[i].getId();
	    ChunkRules.put(id, rules[i]);
	    Lhs lhs = rules[i].getLhs();
	    W2 w2 = lhs.getW2();
	    ChunkRuleW2Index.put(id, w2.getPos());
	    Rhs rhs = rules[i].getRhs();
	    IRhsChoice[] contents = rhs.getContent();
	    for (int j = 0; j < contents.length; j++) {
		if (contents[j] instanceof W1) {
		    W1 w1 = (W1)contents[j];
		    ChunkRuleIndex.put(id, w1.getPos());
		    ChunkRuleIndexLoc.put(id, new Integer(j));
		    break;
		}
	    }
	}
    }

    public static void prepareIPAFormTable() throws Exception {
	// prepare IPAFormTable
	SAXParserFactory spFactory = SAXParserFactory.newInstance();
	SAXParser sParser = spFactory.newSAXParser();
	sParser.parse(chaone.class.getResourceAsStream(ipaform_file_name),
		      new IPAFormTableHandler());
    }

    public static void prepareFPAFormTable() throws Exception {
	// prepare FPAFormTable
	SAXParserFactory spFactory = SAXParserFactory.newInstance();
	SAXParser sParser = spFactory.newSAXParser();
	sParser.parse(chaone.class.getResourceAsStream(fpaform_file_name),
		      new FPAFormTableHandler());
    }

    public static void chaonePreprocess(Element el) {
	// preprocess for unknown POS with English Alphabet
	NodeList nodeList = el.getChildNodes();
	for (int i = 0; i < nodeList.getLength(); i++) {
	    chaonePreprocess4Node(nodeList.item(i));
	}
    }
    
    public static void chaonePreprocess4Node(Node node) {
	// preprocess for unknown POS with English Alphabet for each Node
	if (node.getNodeName().equals("W1")) {
	    Element el = (Element)node;
	    if (el.getAttribute("pos").equals("̤θ")) {
		String str = el.getAttribute("orth");
		Vector vec = new Vector();
		boolean fail = false;
		for (int i = 0; i < str.length(); i++) {
		    String c1 = str.substring(i, i+1);
		    if (EASymbolTable.containsKey(c1)) {
			EASymbol eaSymbol = (EASymbol)EASymbolTable.get(c1);
			vec.add(eaSymbol.createW1());
		    }
		    else {
			fail = true;
		    }
		}
		if (!(fail)) {
		    Node parent = node.getParentNode();
		    for (int i=0; i < vec.size(); i++) {
			Element iel=(Element)vec.elementAt(i);
			parent.insertBefore(iel, node);
		    }
		    parent.removeChild(node);
		} else {
		    String p = el.getAttribute("pron");
		    if (HKTable.containsKey(p)) {
			el.setAttribute("pron", (String)HKTable.get(p));
		    }
		}
	    }
	}
	NodeList nodeList = node.getChildNodes();
	for (int i = 0; i < nodeList.getLength(); i++) {
	    chaonePreprocess4Node(nodeList.item(i));
	}
    }
    
    public static void chaoneChunker(Element el) {
	// Chunker process
	chaoneChunker4Node(el.getChildNodes());
    }
    
    public static void chaoneChunker4Node(NodeList nodeList) {
	// Chunker process for each Node
	int mode = 1; // 1: rule mode, 2: BIE mode
	Vector w1vec = new Vector();
	Node nodeArray[] = new Node[nodeList.getLength()];
	for (int i = 0; i < nodeList.getLength(); i++) {
	    nodeArray[i] = nodeList.item(i);
	}
	for (int i = 0; i < nodeArray.length; i++) {
	    Node node = nodeArray[i];
	    if (debug) {
		System.err.println("node " + i + ": " + node);
	    }
	    if (node.getNodeName().equals("W1")) {
		Element el = (Element)node;
		if (el.getAttribute("w2Chunk").equals("B")) {
		    mode = 2;
		    if (w1vec.size() > 0) {
			chaoneChunkerProcess(w1vec);
			w1vec.clear();
		    }
		}
		w1vec.add(el);
		if (el.getAttribute("w2Chunk").equals("E")) {
		    mode = 1;
		    chaoneChunkerBIEProcess(w1vec);
		    w1vec.clear();
		}
	    }
	    else if (!(node.getNodeName().equals("W2"))) {
		if (node.getNodeType() == Node.ELEMENT_NODE) {
		    mode = 1;
		    if (w1vec.size() > 0) {
			chaoneChunkerProcess(w1vec);
			w1vec.clear();
		    }
		}
		chaoneChunker4Node(node.getChildNodes());
	    }
	}
	if (w1vec.size() > 0) {
	    chaoneChunkerProcess(w1vec);
	    w1vec.clear();
	}
    }
    
    public static void chaoneChunkerProcess(Vector targets) {
	// Chunker for targets
	// select Rules
	Vector rangeList = chaoneChunkerSelectRules(targets);
	// apply rules
	for (int i=0; i < rangeList.size(); i++) {
	    Range range = (Range) rangeList.get(i);
	    if (debug) {
		System.err.println("Chunk Rule: " + range.getId() + " hits from " + targets.get(range.getFrom()) + " to " + targets.get(range.getTo()));
	    }
	    chaoneChunkerApplyRule(range, targets);
	}
    }

    public static Vector chaoneChunkerSelectRules(Vector targets) {
	Vector rangeList = new Vector();
	// calc applicable rules
	for (int i=0; i < targets.size(); i++) {
	    Element el = (Element)targets.get(i);
	    Set set = ChunkRuleIndex.entrySet();
	    Iterator iterator = set.iterator();
	    while(iterator.hasNext()) {
		Map.Entry mapentry = (Map.Entry) iterator.next();
		String val = (String) mapentry.getValue();
		String pos = (String)el.getAttribute("pos");
		if (pos.startsWith(val)) {
		    Range range = chaoneChunkerCheckRule((String)mapentry.getKey(), targets, i);
		    if (range != null)
			rangeList.add(range);
		}
	    }
	}
	// choose compatible ranges
	if (rangeList.size() > 1) {
	    Object rangeArray[] = rangeList.toArray();
	    Arrays.sort(rangeArray, new RangeComparator());
	    rangeList = new Vector();
	    rangeList.add(rangeArray[0]);
	    for (int i = 1; i < rangeArray.length; i++) {
		Range rRange = (Range) rangeArray[i];
		int rfrom = rRange.getFrom();
		int rto = rRange.getTo();
		boolean fail = false;
		for (int j = 0; j < i; j++) {
		    Range qRange = (Range) rangeArray[j];
		    int qfrom = qRange.getFrom();
		    int qto = qRange.getTo();
		    if ((qfrom <= rto) && (qto >= rfrom)) {
			fail = true;
			break;
		    }
		}
		if (!(fail)) {
		    rangeList.add(rRange);
		}
	    }
	}
	return rangeList;
    }
    
    public static Range chaoneChunkerCheckRule(String id, Vector targets, int cloc) {
	// check whether rule is applicable or not
	// if applicable return Range else return null
	Range range = new Range(id, cloc, cloc);
	Rule rule = (Rule)ChunkRules.get(id);
	Rhs rhs = rule.getRhs();
	IRhsChoice[] constraints = rhs.getContent();
	int ruleLoc = ((Integer)ChunkRuleIndexLoc.get(id)).intValue();
	if (debug) {
	    System.err.println("check rule: " + id + " at " + cloc);
	}
	// ruleLoc > 0 : there is pre_cond
	if (ruleLoc > 0) {
	    int from = chaoneChunkerCheckRuleDes(-1, constraints, (ruleLoc - 1), targets, 0, (cloc - 1));
	    if (from != -1) {
		range.setFrom(from);
	    } else {
		return null;
	    }
	}
	// ruleLoc < constraints.length - 1 : there is post_cond
	if (ruleLoc < (constraints.length - 1)) {
	    int to = chaoneChunkerCheckRuleDes(1, constraints, (ruleLoc + 1), targets, (cloc + 1), (targets.size() - 1));
	    if (to != -1) {
		range.setTo(to);
	    } else {
		return null;
	    }
	}
	return range;
    }

    public static int chaoneChunkerCheckRuleDes(int dir, IRhsChoice[] constraints, int ruleLoc, Vector targets, int from, int to) {
	// returns 'from' position, 'to' position or '-1'
	int cur = (dir == -1) ? to : from;
	int org = (dir == -1) ? (to + 1) : (from - 1);
	if (from > to) {
	    boolean fail = false;
	    int cfrom = (dir == -1) ? 0 : ruleLoc;
	    int cto = (dir == -1) ? ruleLoc : (constraints.length - 1);
	    for (int j = cfrom; j < cto; j++) {
		if ((constraints[j] instanceof W1) || (constraints[j] instanceof Choice)) {
		    fail = true;
		    break;
		}
	    }
	    if (fail) {
		return -1;
	    } else {
		return org;
	    }
	}
	Element el = (Element)targets.get(cur);
	String pos = (String)el.getAttribute("pos");
	int nRuleLoc = (dir == -1) ? (ruleLoc - 1) : (ruleLoc + 1);
	int nfrom = (dir == -1) ? from : (from + 1);
	int nto = (dir == -1) ? (to - 1) : to;
	if (constraints[ruleLoc] instanceof W1) {
	    W1 w1 = (W1)constraints[ruleLoc];
	    if (pos.startsWith(w1.getPos())) {
		if ((nRuleLoc < 0) || (nRuleLoc > (constraints.length - 1))) {
		    return cur;
		} else {
		    return chaoneChunkerCheckRuleDes(dir, constraints, nRuleLoc, targets, nfrom, nto);
		}
	    } else {
		return -1;
	    }
	} else if (constraints[ruleLoc] instanceof ZeroOrMore) {
	    ZeroOrMore zom = (ZeroOrMore)constraints[ruleLoc];
	    IZeroOrMoreChoice child = zom.getContent();
	    if (child instanceof W1) {
		W1 w1 = (W1)child;
		if (pos.startsWith(w1.getPos())) {
		    return chaoneChunkerCheckRuleDes(dir, constraints, ruleLoc, targets, nfrom, nto);
		} else {
		    if ((nRuleLoc < 0) || (nRuleLoc > (constraints.length - 1))) {
			return org;
		    } else {
			return chaoneChunkerCheckRuleDes(dir, constraints, nRuleLoc, targets, from, to);
		    }
		}
	    } else if (child instanceof Choice) {
		Choice cho = (Choice)child;
		W1[] w1lis = cho.getW1();
		boolean hit = false;
		for (int j = 0; j < w1lis.length; j++) {
		    if (pos.startsWith(w1lis[j].getPos())) {
			hit = true;
			break;
		    }
		}
		if (hit) {
		    return chaoneChunkerCheckRuleDes(dir, constraints, ruleLoc, targets, nfrom, nto);
		} else {
		    if ((nRuleLoc < 0) || (nRuleLoc > (constraints.length - 1))) {
			return org;
		    } else {
			return chaoneChunkerCheckRuleDes(dir, constraints, nRuleLoc, targets, from, to);
		    }
		}
	    } else {
		return -1;
	    }
	} else if (constraints[ruleLoc] instanceof Optional) {
	    Optional opt = (Optional)constraints[ruleLoc];
	    IOptionalChoice child = opt.getContent();
	    if (child instanceof W1) {
		W1 w1 = (W1)child;
		if (pos.startsWith(w1.getPos())) {
		    if ((nRuleLoc < 0) || (nRuleLoc > (constraints.length - 1))) {
			return cur;
		    } else {
			return chaoneChunkerCheckRuleDes(dir, constraints, nRuleLoc, targets, nfrom, nto);
		    }
		} else {
		    if ((nRuleLoc < 0) || (nRuleLoc > (constraints.length - 1))) {
			return org;
		    } else {
			return chaoneChunkerCheckRuleDes(dir, constraints, nRuleLoc, targets, from, to);
		    }
		}
	    } else if (child instanceof Choice) {
		Choice cho = (Choice)child;
		W1[] w1lis = cho.getW1();
		boolean hit = false;
		for (int j = 0; j < w1lis.length; j++) {
		    if (pos.startsWith(w1lis[j].getPos())) {
			hit = true;
			break;
		    }
		}
		if (hit) {
		    if ((nRuleLoc < 0) || (nRuleLoc > (constraints.length - 1))) {
			return cur;
		    } else {
			return chaoneChunkerCheckRuleDes(dir, constraints, nRuleLoc, targets, nfrom, nto);
		    }
		} else {
		    if ((nRuleLoc < 0) || (nRuleLoc > (constraints.length - 1))) {
			return org;
		    } else {
			return chaoneChunkerCheckRuleDes(dir, constraints, nRuleLoc, targets, from, to);
		    }
		}
	    } else {
		return -1;
	    }
	} else if (constraints[ruleLoc] instanceof Choice) {
	    Choice cho = (Choice)constraints[ruleLoc];
	    W1[] w1lis = cho.getW1();
	    boolean hit = false;
	    for (int j = 0; j < w1lis.length; j++) {
		if (pos.startsWith(w1lis[j].getPos())) {
		    hit = true;
		    break;
		}
	    }
	    if (hit) {
		if ((nRuleLoc < 0) || (nRuleLoc > (constraints.length - 1))) {
		    return cur;
		} else {
		    return chaoneChunkerCheckRuleDes(dir, constraints, nRuleLoc, targets, nfrom, nto);
		}
	    } else {
		return -1;
	    }
	} else { // never enter ...
	    return -1;
	}
    }
    
    public static void chaoneChunkerApplyRule(Range range, Vector targets) {
	String id = range.getId();
	int from = range.getFrom();
	int to = range.getTo();
	Element w2 = doc.createElement("W2");
	Element efrom = (Element)targets.get(from);
	Element eto = (Element)targets.get(to);
	Node parent = efrom.getParentNode();
	parent.insertBefore(w2, efrom);
	String orth = "";
	for (int i = from; i <= to; i++) {
	    Element el = (Element)targets.get(i);
	    orth = orth + el.getAttribute("orth");
	    Node next = el.getNextSibling();
	    if (next instanceof Text) {
		String val = next.getNodeValue();
		if (val.startsWith("\n")) {
		    parent.removeChild(next);
		}
	    }
	    parent.removeChild(el);
	    w2.appendChild(el);
	}
	w2.setAttribute("orth", orth);
	w2.setAttribute("pos", (String)ChunkRuleW2Index.get(id));
	if (eto.getAttributeNode("cType") != null) {
	    w2.setAttribute("cType", eto.getAttribute("cType"));
	}
	if (eto.getAttributeNode("cForm") != null) {
	    w2.setAttribute("cForm", eto.getAttribute("cForm"));
	}
    }

    public static void chaoneChunkerBIEProcess(Vector targets) {
	// Chunker for BIE targets
	Element w2 = doc.createElement("W2");
	Element b = (Element)targets.get(0);
	NamedNodeMap bmap = b.getAttributes();
	for (int j = 0; j < bmap.getLength(); j++) {
	    Attr attr = (Attr)bmap.item(j);
	    String attrname = attr.getNodeName();
	    if (attrname.startsWith("w2_")) {
		w2.setAttribute(attrname.substring(3), attr.getNodeValue());
	    }
	}
	Node parent = b.getParentNode();
	parent.insertBefore(w2, b);
	for (int i=0; i < targets.size(); i++) {
	    Element el = (Element)targets.get(i);
	    NamedNodeMap map = el.getAttributes();
	    for (int j = 0; j < map.getLength(); j++) {
		Attr attr = (Attr)map.item(j);
		String attrname = attr.getNodeName();
		if (attrname.startsWith("w2")) {
		    el.removeAttributeNode(attr);
		}
	    }
	    parent.removeChild(el);
	    w2.appendChild(el);
	}
    }
    
    public static void chaonePAlt(Element el) {
	// ChaOne phonetic alternation routine
	chaonePAlt4Node(el.getChildNodes());
    }

    public static void chaonePAlt4Node(NodeList nodeList) {
	// Phonetic alternation for each node
	// target: W2 with no pron
	for (int i = 0; i < nodeList.getLength(); i++) {
	    Node node = nodeList.item(i);
	    if (node.getNodeName().equals("W2")) {
		Element el = (Element)node;
		if (el.getAttributeNode("pron") == null) {
		    NodeList childList = node.getChildNodes();
		    String pron = "";
		    for (int j = 0; j < childList.getLength(); j++) {
			pron = pron + chaonePAlt4W1(childList.item(j));
		    }
		    // set W2 pron
		    if (debug) {
			System.err.println("W2: " + pron);
		    }
		    el.setAttribute("pron", pron);
		}
	    } else if (!(node.getNodeName().equals("W1"))) {
		chaonePAlt4Node(node.getChildNodes());
	    }
	}
    }
    
    public static String chaonePAlt4W1(Node node) {
	// calc ipaForm, fpaForm, pron for W1 inside W2
	if (node.getNodeName().equals("W1")) {
	    Element el = (Element)node;
	    String ipaType = "";
	    String ipaForm = "";
	    String fpaType = "";
	    String fpaForm = "";
	    String pron = el.getAttribute("pron");
	    if (el.getAttributeNode("ipaType") != null) {
		ipaType = el.getAttribute("ipaType");
		Node prev;
		for (prev = node.getPreviousSibling();
		     prev != null;
		     prev = prev.getPreviousSibling()) {
		    if (prev.getNodeName().equals("W1")) {
			break;
		    }
		}
		if (prev != null) {
		    Element prel = (Element)prev;
		    if (prel.getAttributeNode("ipaConType") != null) {
			ipaForm = (String)IPAFormTable.get(ipaType + ";" + prel.getAttribute("ipaConType"));
			if (ipaForm.length() > 0) {
			    int len = ((String)IPAFormPronTable.get(ipaType + ";ܷ")).length();
			    String alt = (String)IPAFormPronTable.get(ipaType + ";" + ipaForm);
			    pron = alt + pron.substring(len);
			    el.setAttribute("ipaForm", ipaForm);
			}
		    }
		}
	    }
	    if (el.getAttributeNode("fpaType") != null) {
		fpaType = el.getAttribute("fpaType");
		Node post;
		for (post = node.getNextSibling();
		     post != null;
		     post = post.getNextSibling()) {
		    if (post.getNodeName().equals("W1")) {
			break;
		    }
		}
		if (post != null) {
		    Element poel = (Element)post;
		    if (poel.getAttributeNode("fpaConType") != null) {
			String fCT = poel.getAttribute("fpaConType");
			int idx = fCT.indexOf(',');
			if (idx > 0) {
			    fCT = fCT.substring(0, idx);
			}
			fpaForm = (String)FPAFormTable.get(fpaType + ";" + fCT);
			if (fpaForm.length() > 0) {
			    int len = pron.length() - ((String)FPAFormPronTable.get(fpaType + ";ܷ")).length();
			    String alt = (String)FPAFormPronTable.get(fpaType + ";" + fpaForm);
			    pron = pron.substring(0, len) + alt;
			    el.setAttribute("fpaForm", fpaForm);
			}
		    }
		}
	    }
	    el.setAttribute("pron", pron);
	    return pron;
	} else {
	    return "";
	}
    }
    
    public static void printUsage() {
	// show software usage information
	System.err.println("usage: java chaone [-e={ISO-2022-JP|EUC-JP|Shift_JIS}] [-s={chunker|chaone}] [<input_file>]");
	System.exit(-1);
    }
    
    public static void printVersion() {
	// show software version information
	System.err.println("ChaOne version " + version);
	System.exit(-1);
    }
}
