/* Copyright (c) 2000-2003                  */
/*   Yamashita Lab., Ritsumeikan University */
/*   All rights reserved                    */

/*   add preprocess for numbers reading     */
/*                            by Studio ARC */
/*                             2003.08.10   */
/*               version as of 2003.08.18   */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>
#include	<ctype.h>
#include	"synthesis.h"
#include	"confpara.h"
#include	"slot.h"

int TmpMsg(char *,...);
int LogMsg(char *,...);
int ErrMsg(char *,...);
void restart(int);
int make_chasen_process(int *,int *);
int parse_chasen_morph(char *);
int parse_chasen_morph1(char *);
int open_chasen_compound(char *);
void close_chasen_compound();
void parse_tag(char *, int);
void make_sil_morph(SILENCE);
void make_aphrase();
void print_aphrase();
void make_breath();
void print_breath();
void make_phoneme();
void print_phoneme();
void make_sentence();
void print_sentence();
void print_tag();
void modify_morph();
void make_mora();
void print_mora();
void print_morph();

static int fd_in, fd_out;

void init_text_analysis()
{
	int err;
	err = make_chasen_process( &fd_in, &fd_out );
	if( err )  exit(1);
}	

int read_line( int fd, char *buf, int len )
{
	char *buffer;

	buffer = buf;
	while( read( fd, buf, 1 ) != 0 )  {
		if( *buf == '\n' )  {
			*(buf) = '\0';		/* NLɤϺ */
			return( 0 );
		}
		++buf;  --len;
		if( len <= 1 )  {
			ErrMsg( "Too long line ...\n%s\n", buffer );
			*buf = '\0';
			return( 0 );
		}
	}
	*buf = '\0';
	return( EOF );
}	

#define	MAX_TEXT_LINE 2048

extern BOOLEAN in_compound;

void make_morph( char *text )
{
	int n, p;
	char buf[MAX_TEXT_LINE], *line;

	init_text_analysis();

	n = strlen( text );
	if( write( fd_out, text, n ) != n )  {
		ErrMsg( "write error\n" );
		restart( 1 );
	}
	write( fd_out, "\n", 1 );

	in_compound = NO;
	p = 0;		/* Ƭ鲿ʸܤ */
	if( logfp && slot_Log_chasen )  LogMsg( "* chasen result\n" );
	make_sil_morph( SILB );
	while( read_line( fd_in, buf, sizeof(buf) ) != EOF )  {
		if( logfp && slot_Log_chasen )  LogMsg( "%s\n", buf );

		line = buf;
		while( *line == ' ' || *line == '\t' )  ++line;
		if( *line == '\0' )  continue;

		if( strncmp(line,"</S>",4)==0 )  {
			break;
		} else if( strncmp(line,"<S></S>",7)==0 )  {
			/* ʥȡ䥤 <S></S>֤褦*/
			break;
		} else if( strncmp(line,"<S>",3)==0 )  {
			/* do nothing */
		} else if( strncmp(line,"<W ",3)==0 )  {
			p += parse_chasen_morph( line );
		} else if( strncmp(line,"<W1 ",4)==0 )  {
			p += parse_chasen_morph1( line );
		} else if( strncmp(line,"<W2 ",4)==0 )  {
			open_chasen_compound( line );
		} else if( strncmp(line,"</W2>",5)==0 )  {
			close_chasen_compound();
		} else {
			parse_tag( line, p );
		}
	}
	/* ֡פ֡פǽäƤʤ SILE ĤƤʤ*/
	if( mptail->silence != SILE )  make_sil_morph( SILE );
}

#define SEN 0

const char *kansuuji[] = {
  "", "", "", "", "",
  "", "ϻ", "", "Ȭ", "",
};

const char *keta[] = {
  "",
  "", "", "ɴ", "",
  "", "", "ɴ", "",
  "", "", "ɴ", "",
  "", "", "ɴ", "",
};

void zen2han(char *arb) {
  /* if arb is Zenkaku, trans to Hankaku */
  char	*buf,*p,*ptr;
  buf=(char *)calloc(strlen(arb)+1,sizeof(char));
  for(ptr=arb, p=buf; *ptr!='\0'; *ptr++) {
    if (strncmp(ptr,"",2)==0) {*p='!';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='"';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='#';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='$';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='%';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='&';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='\'';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='(';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p=')';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='*';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='+';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p=',';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='-';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='.';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='/';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p=':';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p=';';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='<';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='=';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='>';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='?';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='@';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='[';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='\\';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p=']';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='^';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='_';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='`';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='{';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='|';p++;ptr++;}
    else if (strncmp(ptr,"",2)==0) {*p='}';p++;ptr++;}
    else if ( *ptr == (char)0xA3 ) {
      ptr++;
      if ( *ptr >= (char)0xB0 && *ptr <= (char)0xB9 ) {
	*p = *ptr - 0xB0 + '0';
	p++;
      } else if ( *ptr >= (char)0xC1 && *ptr <= (char)0xDA ) {
	*p = *ptr - 0xC1 + 'A';
	p++;
      } else if ( *ptr >= (char)0xE1 && *ptr <= (char)0xFA ) {
	*p = *ptr - 0xE1 + 'a';
	p++;
      } else { *p=*ptr; p++; }
    } else { *p=*ptr; p++; }
  }
  strcpy(arb, buf);
  free(buf);
}

void arabic2kansuuji(char *arb, char *knj, int kp) {
  int i, kt1, kt2, num, len, flag, zflag;
  *knj = '\0';
  flag = 0;
  len = strlen(arb);
  if ( kp == 1 ) {
    /* ̼ */
    if ( len < sizeof(keta) / sizeof(keta[0]) ) {
      zflag = 1;
      for ( i=0; i<len; i++ ) {
	num = *(arb+i) - '0'; /*  */
	kt1 = len - i; /*  */
	kt2 = kt1 % 4; /* ڤ */
	if ( num && (num > 1 || kt2 == 1 
#if SEN
		     || kt2 == 0
#endif
		     )) {
	  strcat(knj, kansuuji[num]);
	  flag = 1;
	  zflag = 0;
	} else if ( zflag == 1 && num == 0 && kt1 == 1 ) {
	  strcat(knj, kansuuji[num]);
	}
	if (kt1 > 1 && (num || kt2 == 1)) {
	  if (flag || num == 1) {
	    strcat(knj, keta[kt1]);
	    flag = 1;
	    zflag = 0;
	  }
	}
	if (kt2 == 1) flag = 0;
      }
    }
  } else {
    for ( i=0; i<len; i++ ) {
      num = *(arb+i) - '0';
      strcat(knj, kansuuji[num]);
    }
  }
}

void a2k4number (char *cont, char *kcont, char pc, char kc) {
  /* Ⱦѥӥ̼ɤߤѴѴ */
  /* 1234 -> ɴ */
  /* ̼ڤ국(kc)ɤߤȤФ */
  /* 1,234 -> ɴ */
  /* (pc)ɤ */
  /* 12.34 -> 󡥻 */
  /* ʳϤɤߤȤФ */
  char *tpt;
  char ktmp[128];
  char ttmp[32];
  int tcp = 0;
  int fst = 1;
  int p = 0;
  tpt = cont;
  while ( *tpt ) {
    while ( *tpt == kc || *tpt >= '0' && *tpt <= '9' && *tpt != '\0' ) {
      if (*tpt != kc) {
	ttmp[tcp++] = *tpt;
      }
      ++tpt;
    }
    ttmp[tcp] = '\0';
    tcp = 0;
    if ( strcmp(ttmp, "0") == 0 ) { fst = 0; }
    arabic2kansuuji( ttmp, ktmp, fst );
    fst = 0;
    if ( *tpt == pc ) {
      strcat(ktmp, "");
      tpt++;
    } else if (*tpt != '\0') {
      /* ʳɤߤȤФ */
      tpt++;
    }
    strncpy( kcont+p, ktmp, strlen(ktmp) );
    p += strlen(ktmp);
  }
  kcont[p++] = ' ';
  kcont[p] = '\0';
}

void a2k4digit (char *cont, char *kcont) {
  /* ȾѥӥʸѴѴ */
  /* 1234 -> 󻰻 */
  /* ʳϤɤߤȤФ */
  char *tpt;
  char ktmp[128];
  char ttmp[32];
  int tcp = 0;
  int p = 0;
  tpt = cont;
  while ( *tpt ) {
    while ( *tpt >= '0' && *tpt <= '9' && *tpt != '\0' ) {
      ttmp[tcp++] = *tpt;
      ++tpt;
    }
    ttmp[tcp] = '\0';
    tcp = 0;
    arabic2kansuuji( ttmp, ktmp, 0 );
    if (*tpt != '\0') {
      /* ʳɤߤȤФ */
      tpt++;
    }
    strncpy( kcont+p, ktmp, strlen(ktmp) );
    p += strlen(ktmp);
  }
  kcont[p++] = ' ';
  kcont[p] = '\0';
}

void a2k4date (char *cont, char *kcont, char *format, char dlm) {
  /* ǥߥǶڤ줿Ⱦѥӥդ򤢤魯ѴѴ */
  /* 2003-8-3 -> 黰ǯȬ */
  /* Ϳ줿ڤ국(dlm)Ѥ */
  /* 2003/8/3 (with delim = '/') */
  /* formatǯ(Y)(M)(D)ν */
  /* 8-3-2003 (with format = "MDY") */
  /* ʳϤɤߤȤФ */
  /* date˴ؤơŪɤͿ */
  /* 򤬤Ǥޤ */
  /* о: 1,2,3,4,5,6,7,8,9,10,20 */
  char *tpt;
  char ktmp[64];
  char ttmp[32];
  char ytmp[64];
  char mtmp[64];
  char dtmp[128];
  int ord = 0;
  int tcp = 0;
  int p = 0;
  tpt = cont;
  while ( *tpt ) {
    while ( *tpt >= '0' && *tpt <= '9' ) {
      ttmp[tcp++] = *tpt;
      ++tpt;
    }
    if ( *tpt == dlm || *tpt == '\0' ) {
      ttmp[tcp] = '\0';
      tcp = 0;
      if ( *tpt != '\0' ) { tpt++; }
      if ( strlen(format) <= ord ) {
	arabic2kansuuji( ttmp, ktmp, 0 );
      } else if ( format[ord] == 'Y' ) {
	if ( strncmp(ttmp, "0", 1) == 0 ) {
	  arabic2kansuuji( ttmp, ytmp, 0 );
	} else {
	  arabic2kansuuji( ttmp, ytmp, 1 );
	}
	strcat(ytmp, "ǯ");
      } else if ( format[ord] == 'M' ) {
	arabic2kansuuji( ttmp, mtmp, 1 );
	strcat(mtmp, "");
      } else if ( format[ord] == 'D' ) {
	int yomi = 1;
	dtmp[0] = '\0';
	if ( strcmp(ttmp, "1") == 0 ||  strcmp(ttmp, "01") == 0 ) {
	  strcat(dtmp, "<PRON SYM='ĥ'>");
	} else if ( strcmp(ttmp, "2") == 0 ||  strcmp(ttmp, "02") == 0 ) {
	  strcat(dtmp, "<PRON SYM='եĥ'>");
	} else if ( strcmp(ttmp, "3") == 0 ||  strcmp(ttmp, "03") == 0 ) {
	  strcat(dtmp, "<PRON SYM='ߥå'>");
	} else if ( strcmp(ttmp, "4") == 0 ||  strcmp(ttmp, "04") == 0 ) {
	  strcat(dtmp, "<PRON SYM='å'>");
	} else if ( strcmp(ttmp, "5") == 0 ||  strcmp(ttmp, "05") == 0 ) {
	  strcat(dtmp, "<PRON SYM='ĥ'>");
	} else if ( strcmp(ttmp, "6") == 0 ||  strcmp(ttmp, "06") == 0 ) {
	  strcat(dtmp, "<PRON SYM='।'>");
	} else if ( strcmp(ttmp, "7") == 0 ||  strcmp(ttmp, "07") == 0 ) {
	  strcat(dtmp, "<PRON SYM='ʥΥ'>");
	} else if ( strcmp(ttmp, "8") == 0 ||  strcmp(ttmp, "08") == 0 ) {
	  strcat(dtmp, "<PRON SYM='衼'>");
	} else if ( strcmp(ttmp, "9") == 0 ||  strcmp(ttmp, "09") == 0 ) {
	  strcat(dtmp, "<PRON SYM='Υ'>");
	} else if ( strcmp(ttmp, "10") == 0 ) {
	  strcat(dtmp, "<PRON SYM='ȡ'>");
	} else if ( strcmp(ttmp, "20") == 0 ) {
	  strcat(dtmp, "<PRON SYM='ϥĥ'>");
	} else {
	  yomi = 0;
	}
	arabic2kansuuji( ttmp, ktmp, 1 );
	strcat(dtmp, ktmp);
	strcat(dtmp, "");
	if ( yomi == 1 ) {
	  strcat(dtmp, "</PRON>");
	}
      }
      ord++;
    } else if (*tpt != '\0') {
      /* ʳɤߤȤФ */
      tpt++;
    }
  }
  strncpy( kcont+p, ytmp, strlen(ytmp) );
  p += strlen(ytmp);
  strncpy( kcont+p, mtmp, strlen(mtmp) );
  p += strlen(mtmp);
  strncpy( kcont+p, dtmp, strlen(dtmp) );
  p += strlen(dtmp);
  kcont[p++] = ' ';
  kcont[p] = '\0';
}

void a2k4time (char *cont, char *kcont, char *format, char dlm) {
  /* ǥߥǶڤ줿Ⱦѥӥ򤢤魯ѴѴ */
  /* 12:34:56 -> ʬ޽ϻ */
  /* Ϳ줿ڤ국(dlm)Ѥ */
  /* 12/34/56 (with delim = '/') */
  /* formatǻ(h)ʬ(m)(s) */
  /* 12:34 (with format = "hm") */
  /* ʳϤɤߤȤФ */
  char *tpt;
  char ktmp[128];
  char ttmp[32];
  int ord = 0;
  int tcp = 0;
  int p = 0;
  tpt = cont;
  while ( *tpt ) {
    while ( *tpt >= '0' && *tpt <= '9' ) {
      ttmp[tcp++] = *tpt;
      ++tpt;
    }
    if ( *tpt == dlm || *tpt == '\0' ) {
      ttmp[tcp] = '\0';
      tcp = 0;
      if ( *tpt != '\0' ) { tpt++; }
      if ( strlen(format) <= ord ) {
	arabic2kansuuji( ttmp, ktmp, 0 );
      } else if ( format[ord] == 'h' ) {
	arabic2kansuuji( ttmp, ktmp, 1 );
	strcat(ktmp, "");
      } else if ( format[ord] == 'm' ) {
	arabic2kansuuji( ttmp, ktmp, 1 );
	strcat(ktmp, "ʬ");
      } else if ( format[ord] == 's' ) {
	arabic2kansuuji( ttmp, ktmp, 1 );
	strcat(ktmp, "");
      }
      ord++;
      strncpy( kcont+p, ktmp, strlen(ktmp) );
      p += strlen(ktmp);
    } else if (*tpt != '\0') {
      /* ʳɤߤȤФ */
      tpt++;
    }
  }
  kcont[p++] = ' ';
  kcont[p] = '\0';
}

void a2k4phone (char *cont, char *kcont) {
  /* ֹ򤢤魯ȾѥӥʸѴѴ */
  /* (123)456-7890 -> 󻰡͸ϻȬ塻 */
  /* ǥߥϡѴ */
  /* ʳϤɤߤȤФ */
  char *tpt;
  char ktmp[128];
  char ttmp[32];
  int tcp = 0;
  int p = 0;
  tpt = cont;
  if ( *tpt == '(' ) { tpt++; }
  while ( *tpt ) {
    while ( *tpt >= '0' && *tpt <= '9' && *tpt != '\0' ) {
      ttmp[tcp++] = *tpt;
      ++tpt;
    }
    ttmp[tcp] = '\0';
    tcp = 0;
    arabic2kansuuji( ttmp, ktmp, 0 );
    if ( *tpt == '(' || *tpt == ')' || *tpt == '-' ) {
      strcat(ktmp, "");
      tpt++;
    } else if (*tpt != '\0') {
      /* ʳɤߤȤФ */
      tpt++;
    }
    strncpy( kcont+p, ktmp, strlen(ktmp) );
    p += strlen(ktmp);
  }
  kcont[p++] = ' ';
  kcont[p] = '\0';
}

void spell_process (char *cont, char *kcont) {
  /* ʸʸĥڡڤꤵ줿ʸѴ */
  /* ABC ->    */
  char *tpt;
  char ktmp[128];
  int p = 0;
  int c = 0;
  tpt = cont;
  while ( *tpt ) {
    if ( *tpt >= '0' && *tpt <= '9' ) {
      kcont[p++] = 0xA3;
      kcont[p++] = 0xB0 + *tpt - '0';
      kcont[p++] = ' ';
    } else if( 'A' <= *tpt && *tpt <= 'Z' )  {
      kcont[p++] = 0xA3;
      kcont[p++] = 0xC1 + *tpt - 'A';
      kcont[p++] = ' ';
    } else if( 'a' <= *tpt && *tpt <= 'z' )  {
      kcont[p++] = 0xA3;
      kcont[p++] = 0xE1 + *tpt - 'a';
      kcont[p++] = ' ';
    } else if( *tpt == '!' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '"' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '#' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '$' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '%' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '&' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '\'' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '(' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == ')' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '*' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '+' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == ',' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '-' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '.' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '/' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '=' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == '?' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == ':' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else if( *tpt == ';' )  {
      strncpy( kcont+p, "", 2 );  p += 2;
    } else {
      kcont[p++] = *(tpt);
      if (c == 0) {
	c++;
      } else {
	c = 0;
	kcont[p++] = ' ';
      }
    }
    ++tpt;
  }
  kcont[p++] = ' ';
  kcont[p] = '\0';
}


void flushnbuf( char *buf, char *kcont, int nmode )
{
  if ( nmode == 1 ) { /* NUMBER mode */
    if ( strcmp(read_number, "DECIMAL") == 0 ) {
      a2k4number(buf, kcont, '.', ',');
    } else {
      a2k4digit(buf, kcont);
    }
  } else if ( nmode == 2 ) { /* DATE mode */
    if ( strcmp(read_date, "NO") == 0 ) {
      if ( strcmp(read_number, "DECIMAL") == 0 ) {
	a2k4number(buf, kcont, '.', '-');
      } else {
	a2k4digit(buf, kcont);
      }
    } else {
      a2k4date(buf, kcont, read_date, '-');
    }
  } else if ( nmode == 3 ) { /* TIME mode */
    if ( strcmp(read_time, "NO") == 0 ) {
      if ( strcmp(read_number, "DECIMAL") == 0 ) {
	a2k4number(buf, kcont, '.', '-');
      } else {
	a2k4digit(buf, kcont);
      }
    } else {
      a2k4time(buf, kcont, read_time, ':');
    }
  } else if ( nmode == 4 ) { /* ALPHABET mode */
    if ( strcmp(read_alphabet, "NO") == 0 ) {
      spell_process(buf, kcont);
    } else {
      strncpy(kcont, buf, strlen(buf));
      kcont[strlen(buf)] = '\0';
    }
  }
}

/* ȾѥڡνȾʸʸѴ*/
/* context˴ؤ (by Studio ARC 2003.08.03) */
void arrange_text( char *text, char *utterance )
{
	int p, tp, cp, in_context, in_spell;
	char buf[MAX_TEXT_LINE];
	char type[32];
	char format[32];
	char delim[32];
	char cont[64];
	char kcont[128];
	char tcont[32];
	int nmode = 0;

	p = tp = cp = 0;
	in_context = in_spell = 0;

	while( *text )  {
	  if( *text == '<' )  { /* λϤޤ */
	    /* nmodeν */
	    if ( nmode > 0 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
	    }
	    tp = 0;
	    buf[tp++] = *text;
	    ++text;
	    if ( *text == '/' ) { /* λλϤޤ */
	      buf[tp++] = *text;
	      ++text;
	      while ( *text != '>' ) {
		buf[tp++] = *text;
/*		buf[tp++] = toupper( *text );  /* ̾ʸѴ */
		++text;
	      }
	      buf[tp++] = *text;
	      buf[tp] = '\0';
	      if ( strncmp(buf, "</CONTEXT>", 10) == 0) {
		if ( in_context == 1 ) { /* in_contextνλ */
		  cont[cp] = '\0';
		  zen2han(cont);
		  if ( strcmp(type, "NUMBER") == 0 ) {
		    if ( strcmp(format, "ISO") == 0 ) {
		      a2k4number(cont, kcont, ',', ' ');
		    } else {
		      a2k4number(cont, kcont, '.', ',');
		    }
		    strncpy( utterance+p, kcont, strlen(kcont) );
		    p += strlen(kcont);
		  } else if ( strcmp(type, "DIGITS") == 0 ) {
		    a2k4digit(cont, kcont);
		    strncpy( utterance+p, kcont, strlen(kcont) );
		    p += strlen(kcont);
		  } else if ( strcmp(type, "DATE") == 0 ) {
		    char dlm;
		    if ( strlen(delim) > 0 ) {
		      dlm = delim[0];
		    } else {
		      dlm = '-';
		    }
		    if ( strlen(format) > 0 ) {
		      a2k4date(cont, kcont, format, dlm);
		    } else {
		      a2k4date(cont, kcont, "YMD", dlm);
		    }
		    strncpy( utterance+p, kcont, strlen(kcont) );
		    p += strlen(kcont);
		  } else if ( strcmp(type, "TIME") == 0 ) {
		    char dlm;
		    if ( strlen(delim) > 0 ) {
		      dlm = delim[0];
		    } else {
		      dlm = ':';
		    }
		    if ( strlen(format) > 0 ) {
		      a2k4time(cont, kcont, format, dlm);
		    } else {
		      a2k4time(cont, kcont, "hms", dlm);
		    }
		    strncpy( utterance+p, kcont, strlen(kcont) );
		    p += strlen(kcont);
		  } else if ( strcmp(type, "PHONE") == 0 ) {
		    a2k4phone(cont, kcont);
		    strncpy( utterance+p, kcont, strlen(kcont) );
		    p += strlen(kcont);
		  }
		  tp = cp = 0;
		  in_context = 0;
		} else { /* ERROR: CONTEXTλˤ⤫餺in-contextǤʤ */
		  strncpy( utterance+p, buf, strlen(buf) );
		  p += strlen(buf);
		  tp = cp = 0;
		}
	      } else if ( strncmp(buf, "</SPELL>", 8) == 0) {
		if ( in_spell == 1 ) { /* in_spellνλ */
		  cont[cp] = '\0';
		  spell_process(cont, kcont);
		  strncpy( utterance+p, kcont, strlen(kcont) );
		  p += strlen(kcont);
		  in_spell = 0;
		} else { /* ERROR: SPELLλˤ⤫餺in-spellǤʤ */
		  strncpy( utterance+p, buf, strlen(buf) );
		  p += strlen(buf);
		  tp = cp = 0;
		}
	      } else { /* CONTEXT, SPELLʳνλν */
		strncpy( utterance+p, buf, strlen(buf) );
		p += strlen(buf);
		tp = cp = 0;
	      }
	    } else { /* ϥλϤޤ */
	      while ( *text != '>' ) {
		buf[tp++] = *text;
/*		buf[tp++] = toupper( *text );  /* ʸѴ */
		++text;
	      }
	      buf[tp++] = *text;
	      buf[tp] = '\0';
	      if ( strncmp(buf, "<CONTEXT", 8) == 0) {
		type[0] = '\0';
		format[0] = '\0';
		delim[0] = '\0';
		if ( strstr(buf, "TYPE=\"") != NULL ) {
		  char *pb;
		  pb = strstr(buf, "TYPE=\"");
		  pb = pb + 6;
		  while (*pb != '"') {
		    type[cp++] = *pb;
		    ++pb;
		  }
		  type[cp] = '\0';
		  cp = 0;
		  in_context = 1;
		} else {
		  strncpy( utterance+p, buf, strlen(buf) );
		  p += strlen(buf);
		  tp = cp = 0;
		}
		if ( strstr(buf, "FORMAT=\"") != NULL ) {
		  char *pb;
		  pb = strstr(buf, "FORMAT=\"");
		  pb = pb + 8;
		  while (*pb != '"') {
		    format[cp++] = *pb;
		    ++pb;
		  }
		  format[cp] = '\0';
		  cp = 0;
		}
		if ( strstr(buf, "DELIM=\"") != NULL ) {
		  char *pb;
		  pb = strstr(buf, "DELIM=\"");
		  pb = pb + 7;
		  while (*pb != '"') {
		    delim[cp++] = *pb;
		    ++pb;
		  }
		  delim[cp] = '\0';
		  cp = 0;
		}
	      } else if ( strncmp(buf, "<SPELL", 6) == 0) { /* SPELL */
		tp = cp = 0;
		in_spell = 1;
	      } else { /* CONTEXT, SPELLʳγϥ */
		strncpy( utterance+p, buf, strlen(buf) );
		p += strlen(buf);
		tp = cp = 0;
	      }
	    }
	  } else if ( in_context == 1 || in_spell == 1) { /* context, spellΥƥ */
	    cont[cp++] = *text;
	  } else if (strncmp(text,"",2)==0) {
	    if ( nmode == 4 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
	    }
	    if ( nmode == 1 ) {
	      buf[tp++] = '.';
	    } else {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    }
	    text++;
	  } else if (strncmp(text,"",2)==0) {
	    if ( nmode == 4 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
	    }
	    if ( nmode == 1 || nmode == 2 ) {
	      buf[tp++] = '-';
	    } else {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    }
	    text++;
	  } else if (strncmp(text,"",2)==0) {
	    if ( nmode == 4 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
	    }
	    if ( nmode == 1 || nmode == 3 ) {
	      buf[tp++] = ':';
	    } else {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    }
	    text++;
	  } else if ( *text == '.' )  {
	    if ( nmode == 4 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
	    }
	    if ( nmode == 1 ) {
	      buf[tp++] = *text;
	    } else {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    }
	  } else if ( *text == '-' )  {
	    if ( nmode == 4 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
	    }
	    if ( nmode == 1 || nmode == 2 ) {
	      buf[tp++] = *text;
	      nmode = 2;
	    } else {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    }
	  } else if ( *text == ':' )  {
	    if ( nmode == 4 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
	    }
	    if ( nmode == 1 || nmode == 3 ) {
	      buf[tp++] = *text;
	      nmode = 3;
	    } else {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    }
	  } else if ( *text == (char)0xA3 ) {
	    text++;
	    if ( *text >= (char)0xB0 && *text <= (char)0xB9 ) {
	      char han;
	      han = *text - 0xB0 + '0';
	      if ( nmode == 4 ) {
		buf[tp] = '\0';
		flushnbuf(buf, kcont, nmode);
		strncpy( utterance+p, kcont, strlen(kcont) );
		p += strlen(kcont);
		nmode = tp = 0;
	      }
	      buf[tp++] = han;
	      if ( nmode == 0 ) { nmode = 1; }
	    } else if ( *text >= (char)0xC1 && *text <= (char)0xDA ) {
	      if ( nmode > 0 && nmode < 4 ) {
		buf[tp] = '\0';
		flushnbuf(buf, kcont, nmode);
		strncpy( utterance+p, kcont, strlen(kcont) );
		p += strlen(kcont);
		nmode = tp = 0;
	      }
	      buf[tp++] = (char)0xA3;
	      buf[tp++] = *text;
	      if ( nmode == 0 ) { nmode = 4; }
	    } else if ( *text >= (char)0xE1 && *text <= (char)0xFA ) {
	      if ( nmode > 0 && nmode < 4 ) {
		buf[tp] = '\0';
		flushnbuf(buf, kcont, nmode);
		strncpy( utterance+p, kcont, strlen(kcont) );
		p += strlen(kcont);
		nmode = tp = 0;
	      }
	      buf[tp++] = (char)0xA3;
	      buf[tp++] = *text;
	      if ( nmode == 0 ) { nmode = 4; }
	    } else {
	      utterance[p++] = (char)0xA3;
	      //text--;
	      utterance[p++] = *(text);
	    }
	  } else if ( *text >= (char)0xa1 && *text <= (char)0xf4 ) {
	    /* 2Хʸ */
	    if ( nmode > 0 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      /* ֿ͡פspaceϤȤ */
	      if ( *text == (char)0xbf && *(text + 1) == (char)0xcd ) {
		kcont[strlen(kcont) - 1] = '\0';
	      }
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
 	    }
	    if ( *text == (char)0xa1 && *( text + 1 ) == (char)0xa1 ) {
	      // ѥڡν
	      text++;
	    } else {
	      utterance[p++] = *text++;
	      utterance[p++] = *text;
	    }
	  } else if ( '0' <= *text && *text <= '9' ) {
	    if ( nmode == 4 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
	    }
	    buf[tp++] = *text;
	    if ( nmode == 0 ) { nmode = 1; }
	  } else if ( 'A' <= *text && *text <= 'Z' )  {
	    if ( nmode > 0 && nmode < 4 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
	    }
	    buf[tp++] = 0xA3;
	    buf[tp++] = 0xC1 + *text - 'A';
	    if ( nmode == 0 ) { nmode = 4; }
	  } else if ( 'a' <= *text && *text <= 'z' )  {
	    if ( nmode > 0 && nmode < 4 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
	    }
	    buf[tp++] = 0xA3;
	    buf[tp++] = 0xE1 + *text - 'a';
	    if ( nmode == 0 ) { nmode = 4; }
	  } else {
	    if ( nmode > 0 ) {
	      buf[tp] = '\0';
	      flushnbuf(buf, kcont, nmode);
	      strncpy( utterance+p, kcont, strlen(kcont) );
	      p += strlen(kcont);
	      nmode = tp = 0;
	    }
	    if ( *text == '!' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '"' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '#' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '$' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '%' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '&' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '\'' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '(' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == ')' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '*' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '+' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == ',' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '/' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '=' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == '?' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	    } else if ( *text == ';' )  {
	      strncpy( utterance+p, "", 2 );  p += 2;
	      /* ڡɤФŬϰϤʸǷ뤿ᡣ */
	    } else if ( *text == ' ' )  {
	      /* ¾ʸϤΤޤޥԡ */
	    } else {
	      utterance[p++] = *(text);
	    }
	  }
	  ++text;
	} /* whileʸΤ */
	if ( nmode > 0 ) {
	  buf[tp] = '\0';
	  flushnbuf(buf, kcont, nmode);
	  strncpy( utterance+p, kcont, strlen(kcont) );
	  p += strlen(kcont);
	  nmode = tp = 0;
	}
	utterance[p] = '\0';
}

void text_analysis( char *text )
{
	if( logfp && slot_Log_text )  {
		LogMsg( "* text\n" );
		LogMsg( "%s\n", text );
	}

	arrange_text( text, chasen_input );

	if( logfp && slot_Log_arranged_text )  {
		LogMsg( "* arranged_text\n" );
		LogMsg( "%s\n", chasen_input );
	}

	/* ǲ */
	make_morph( chasen_input );
	if( logfp && slot_Log_tag )  print_tag();

	/* ɤߤν */
	modify_morph();

	/* ⡼ǡκ */
	make_mora();
	if( logfp && slot_Log_morph )  print_morph();
	if( logfp && slot_Log_mora )  print_mora();

	/* ȶκ */
	make_aphrase();

	/* Ƶκ */
	make_breath();
	if( logfp && slot_Log_aphrase )  print_aphrase();

	/* Ƿη */
	make_phoneme();
	if( logfp && slot_Log_phoneme )  print_phoneme();

	/* ʸϤκ */
	make_sentence();
	if( logfp && slot_Log_breath )  print_breath();
	if( logfp && slot_Log_sentence )  print_sentence();
}

/*--------------------------------------------------------------*/
/* ȯ */

/* ե뤫䥤βϷ̤ɤ߹ */
void read_morph( char *ifile )
{
	FILE *fp;
	char buf[MAX_TEXT_LINE], *line;
	int p;

	fp = fopen( ifile, "r" );
	if( fp == NULL )  {
		ErrMsg( "* Can't open ... %s\n", ifile );
		restart( 1 );
	}
	in_compound = NO;
	p = 0;		/* Ƭ鲿ʸܤ */
	if( logfp && slot_Log_chasen )  LogMsg( "* chasen result\n" );
	make_sil_morph( SILB );
	while( fgets( buf, MAX_TEXT_LINE, fp ) != NULL )  {
		if( buf[strlen(buf)-1] == '\n' )  {
			buf[strlen(buf)-1] = '\0';		/* ԥɤ */
		}
        if( logfp && slot_Log_chasen )  LogMsg( "%s\n", buf );

		line = buf;
		while( *line == ' ' || *line == '\t' )  ++line;
		if( *line == '\0' )  continue;

		if( strncmp(line,"<S>",3)==0 )  {
			/* do nothing */
		} else if( strncmp(line,"</S>",4)==0 )  {
			break;
		} else if( strncmp(line,"<W ",3)==0 )  {
			p += parse_chasen_morph( line );
		} else if( strncmp(line,"<W1 ",4)==0 )  {
			p += parse_chasen_morph1( line );
		} else if( strncmp(line,"<W2 ",4)==0 )  {
			open_chasen_compound( line );
		} else if( strncmp(line,"</W2>",5)==0 )  {
			close_chasen_compound();
		} else {
			parse_tag( line, p );
		}
	}
	fclose( fp );

	/* ֡פ֡פǽäƤʤ SILE ĤƤʤ*/
	if( mptail->silence != SILE )  make_sil_morph( SILE );
}

void text_analysis_file( char *file )
{
	/* ǲϷ̤ɤ߹ */
	read_morph( file );
	if( logfp && slot_Log_tag )  print_tag();

	/* ɤߤν */
	modify_morph();

	/* ⡼ǡκ */
	make_mora();
	if( logfp && slot_Log_morph )  print_morph();
	if( logfp && slot_Log_mora )  print_mora();

	/* ȶκ */
	make_aphrase();

	/* Ƶκ */
	make_breath();
	if( logfp && slot_Log_aphrase )  print_aphrase();

	/* Ƿη */
	make_phoneme();
	if( logfp && slot_Log_phoneme )  print_phoneme();

	/* ʸϤκ */
	make_sentence();
	if( logfp && slot_Log_breath )  print_breath();
	if( logfp && slot_Log_sentence )  print_sentence();
}
